package dev.fyloz.trial.colorrecipesexplorer.core.model

import com.fasterxml.jackson.annotation.JsonIgnore
import com.fasterxml.jackson.annotation.JsonProperty
import org.hibernate.annotations.Fetch
import org.hibernate.annotations.FetchMode
import org.springframework.security.core.GrantedAuthority
import org.springframework.security.core.authority.SimpleGrantedAuthority
import java.time.LocalDateTime
import javax.persistence.*
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotNull
import javax.validation.constraints.Size
import kotlin.jvm.Transient


private const val EMPLOYEE_ID_NULL_MESSAGE = "Un numéro d'employé est requis"
private const val EMPLOYEE_LAST_NAME_EMPTY_MESSAGE = "Un nom  est requis"
private const val EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE = "Un prénom est requis"
private const val EMPLOYEE_PASSWORD_EMPTY_MESSAGE = "Un mot de passe est requis"
private const val EMPLOYEE_PASSWORD_TOO_SHORT_MESSAGE = "Le mot de passe doit contenir au moins 8 caractères"

@Entity
data class Employee(
        @Id
        @field:NotNull(message = EMPLOYEE_ID_NULL_MESSAGE)
        override val id: Long,

        val firstName: String = "",

        val lastName: String = "",

        @JsonIgnore
        val password: String = "",

        @JsonIgnore
        val isDefaultGroupUser: Boolean = false,

        @JsonIgnore
        val isSystemUser: Boolean = false,

        @field:ManyToOne
        @Fetch(FetchMode.SELECT)
        var group: EmployeeGroup? = null,

        @Enumerated(EnumType.STRING)
        @ElementCollection(fetch = FetchType.EAGER)
        @Fetch(FetchMode.SUBSELECT)
        @get:JsonIgnore
        val permissions: MutableList<EmployeePermission> = mutableListOf(),

        val lastLoginTime: LocalDateTime? = null
) : IModel {
    @JsonProperty("permissions")
    fun getFlattenedPermissions(): Iterable<EmployeePermission> = getPermissions()
}

/** DTO for creating employees. Allow a [password] a [groupId]. */
data class EmployeeDto(
        @field:NotNull(message = EMPLOYEE_ID_NULL_MESSAGE)
        val id: Long,

        @field:NotBlank(message = EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE)
        val firstName: String,

        @field:NotBlank(message = EMPLOYEE_LAST_NAME_EMPTY_MESSAGE)
        val lastName: String,

        @field:NotBlank(message = EMPLOYEE_PASSWORD_EMPTY_MESSAGE)
        @field:Size(min = 8, message = EMPLOYEE_PASSWORD_TOO_SHORT_MESSAGE)
        val password: String,

        @field:ManyToOne
        @Fetch(FetchMode.SELECT)
        var groupId: Long? = null,

        @Enumerated(EnumType.STRING)
        @ElementCollection(fetch = FetchType.EAGER)
        val permissions: MutableList<EmployeePermission> = mutableListOf()
)

private const val GROUP_NAME_NULL_MESSAGE = "Un nom est requis"
private const val GROUP_PERMISSIONS_EMPTY_MESSAGE = "Au moins une permission est requise"

@Entity
data class EmployeeGroup(
        @Id
        @GeneratedValue(strategy = GenerationType.SEQUENCE)
        override val id: Long? = null,

        @field:NotBlank(message = GROUP_NAME_NULL_MESSAGE)
        @field:Size(min = 3)
        @Column(unique = true)
        val name: String = "",

        @field:Size(min = 1, message = GROUP_PERMISSIONS_EMPTY_MESSAGE)
        @Enumerated(EnumType.STRING)
        @ElementCollection(fetch = FetchType.EAGER)
        val permissions: MutableList<EmployeePermission> = mutableListOf(),

        @OneToMany
        @JsonIgnore
        val employees: MutableList<Employee> = mutableListOf()
) : IModel {
    fun getEmployeeCount() = employees.size
}


data class EmployeeLoginRequest(val id: Long, val password: String)


enum class EmployeePermission(val impliedPermissions: List<EmployeePermission> = listOf()) {
    // View
    VIEW_EMPLOYEE,
    VIEW_EMPLOYEE_GROUP,
    VIEW(listOf(

    )),

    // Edit
    EDIT_EMPLOYEE(listOf(VIEW_EMPLOYEE)),
    EDIT_EMPLOYEE_PASSWORD(listOf(EDIT_EMPLOYEE)),
    EDIT_EMPLOYEE_GROUP(listOf(VIEW_EMPLOYEE_GROUP)),
    EDIT(listOf(
            VIEW
    )),

    // Remove
    REMOVE_EMPLOYEE(listOf(EDIT_EMPLOYEE)),
    REMOVE_EMPLOYEE_GROUP(listOf(EDIT_EMPLOYEE_GROUP)),
    REMOVE(listOf(
            EDIT
    )),

    // Others
    SET_BROWSER_DEFAULT_GROUP(listOf(
            VIEW_EMPLOYEE_GROUP
    )),

    ADMIN(listOf(
            REMOVE,
            SET_BROWSER_DEFAULT_GROUP,

            // Admin only permissions
            REMOVE_EMPLOYEE,
            EDIT_EMPLOYEE_PASSWORD,
            REMOVE_EMPLOYEE_GROUP,
    ));

    operator fun contains(permission: EmployeePermission): Boolean {
        return permission == this || impliedPermissions.any { permission in it }
    }
}

/** Gets [GrantedAuthority]s of the given [Employee]. */
fun Employee.getAuthorities(): MutableCollection<GrantedAuthority> {
    return getPermissions().map { it.toAuthority() }.toMutableSet()
}

/** Gets [EmployeePermission]s of the given [Employee]. */
fun Employee.getPermissions(): Iterable<EmployeePermission> {
    val grantedPermissions: MutableSet<EmployeePermission> = mutableSetOf()
    if (group != null) grantedPermissions.addAll(group!!.permissions.flatMap { it.flat() })
    grantedPermissions.addAll(permissions.flatMap { it.flat() })
    return grantedPermissions
}

private fun EmployeePermission.flat(): Iterable<EmployeePermission> {
    return mutableSetOf(this).apply {
        impliedPermissions.forEach {
            addAll(it.flat())
        }
    }
}

/** Converts the given [EmployeePermission] to a [GrantedAuthority]. */
private fun EmployeePermission.toAuthority(): GrantedAuthority {
    return SimpleGrantedAuthority(name)
}
