package dev.fyloz.trial.colorrecipesexplorer.core.services

import dev.fyloz.trial.colorrecipesexplorer.core.configuration.SecurityConfigurationProperties
import dev.fyloz.trial.colorrecipesexplorer.core.configuration.blacklistedJwtTokens
import dev.fyloz.trial.colorrecipesexplorer.core.configuration.defaultGroupCookieName
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.*
import dev.fyloz.trial.colorrecipesexplorer.core.model.*
import dev.fyloz.trial.colorrecipesexplorer.dao.EmployeeGroupRepository
import dev.fyloz.trial.colorrecipesexplorer.dao.EmployeeRepository
import org.springframework.beans.factory.annotation.Autowired
import org.springframework.security.core.userdetails.User
import org.springframework.security.core.userdetails.UserDetails
import org.springframework.security.core.userdetails.UserDetailsService
import org.springframework.security.core.userdetails.UsernameNotFoundException
import org.springframework.security.crypto.password.PasswordEncoder
import org.springframework.stereotype.Service
import org.springframework.web.util.WebUtils
import java.time.LocalDateTime
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse
import javax.transaction.Transactional

@Service
class EmployeeService(val employeeRepository: EmployeeRepository, val passwordEncoder: PasswordEncoder) :
        AbstractModelService<Employee, EmployeeRepository>(employeeRepository, Employee::class.java) {
    @Autowired
    lateinit var groupService: EmployeeGroupService

    /** Check if an [Employee] with the given [firstName] and [lastName] exists. */
    fun existsByFirstNameAndLastName(firstName: String, lastName: String): Boolean {
        return repository.existsByFirstNameAndLastName(firstName, lastName)
    }

    override fun getAll(): Collection<Employee> {
        return super.getAll().filter { !it.isSystemUser && !it.isDefaultGroupUser }
    }

    override fun getById(id: Long): Employee {
        return getById(id, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)
    }

    /** Gets the employee with the given [id]. */
    fun getById(id: Long, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): Employee {
        return super.getById(id).apply {
            if (ignoreSystemUsers && isSystemUser || ignoreDefaultGroupUsers && isDefaultGroupUser) throw EntityNotFoundRestException(id)
        }
    }

    /** Gets all employees which have the given [group]. */
    fun getByGroup(group: EmployeeGroup): Collection<Employee> {
        return repository.findByGroup(group).filter { !it.isSystemUser && !it.isDefaultGroupUser }
    }

    /** Gets the default user of the given [group]. */
    fun getDefaultGroupUser(group: EmployeeGroup): Employee {
        return repository.findByIsDefaultGroupUserIsTrueAndGroupIs(group)
    }

    /** Saves the given [employee]. The password contained in the DTO will be hashed in the created [Employee]. */
    fun save(employee: EmployeeDto): Employee {
        return save(with(employee) {
            Employee(id, firstName, lastName, passwordEncoder.encode(password), isDefaultGroupUser = false, isSystemUser = false, group = if (groupId != null) groupService.getById(groupId!!) else null, permissions = permissions)
        })
    }

    override fun save(entity: Employee): Employee {
        if (existsByFirstNameAndLastName(entity.firstName, entity.lastName))
            throw EntityAlreadyExistsException(type, ModelException.IdentifierType.NAME, "${entity.firstName} ${entity.lastName}")

        return super.save(entity)
    }

    /** Updates the last login time of the employee with the given [employeeId]. */
    fun updateLastLoginTime(employeeId: Long) {
        update(Employee(id = employeeId, lastLoginTime = LocalDateTime.now()), ignoreDefaultGroupUsers = true, ignoreSystemUsers = false)
    }

    override fun update(entity: Employee): Employee {
        return update(entity, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)
    }

    /** Updates de given [entity]. **/
    fun update(entity: Employee, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): Employee {
        val persistedEmployee = getById(entity.id, ignoreDefaultGroupUsers, ignoreSystemUsers)
        with(repository.findByFirstNameAndLastName(entity.firstName, entity.lastName)) {
            if (this != null && id != entity.id)
                throw EntityAlreadyExistsRestException("${entity.firstName} ${entity.lastName}")
        }

        return super.update(with(entity) {
            Employee(
                    id,
                    if (firstName.isNotBlank()) firstName else persistedEmployee.firstName,
                    if (lastName.isNotBlank()) lastName else persistedEmployee.lastName,
                    persistedEmployee.password,
                    if (ignoreDefaultGroupUsers) false else persistedEmployee.isDefaultGroupUser,
                    if (ignoreSystemUsers) false else persistedEmployee.isSystemUser,
                    persistedEmployee.group,
                    if (permissions.isNotEmpty()) permissions else persistedEmployee.permissions,
                    lastLoginTime ?: persistedEmployee.lastLoginTime
            )
        })
    }

    /** Updates the password of the employee with the given [id]. */
    fun updatePassword(id: Long, password: String) {
        val persistedEmployee = getById(id, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)
        super.update(with(persistedEmployee) {
            Employee(
                    id,
                    firstName,
                    lastName,
                    passwordEncoder.encode(password),
                    isDefaultGroupUser,
                    isSystemUser,
                    group,
                    permissions,
                    lastLoginTime
            )
        })
    }

    /** Adds the given [permission] to the employee with the given [employeeId]. */
    fun addPermission(employeeId: Long, permission: EmployeePermission) = super.update(getById(employeeId).apply { permissions += permission })

    /** Removes the given [permission] from the employee with the given [employeeId]. */
    fun removePermission(employeeId: Long, permission: EmployeePermission) = super.update(getById(employeeId).apply { permissions -= permission })

    /** Logout an user. Add the authorization token of the given [request] to the blacklisted tokens. */
    fun logout(request: HttpServletRequest) {
        val authorizationCookie = WebUtils.getCookie(request, "Authorization")
        if (authorizationCookie != null) {
            val authorizationToken = authorizationCookie.value
            if (authorizationToken != null && authorizationToken.startsWith("Bearer")) {
                blacklistedJwtTokens.add(authorizationToken)
            }
        }
    }
}

const val defaultGroupCookieMaxAge = 10 * 365 * 24 * 60 * 60 // 10 ans

@Service
class EmployeeGroupService(val employeeGroupRepository: EmployeeGroupRepository, val employeeService: EmployeeService) : AbstractModelService<EmployeeGroup, EmployeeGroupRepository>(employeeGroupRepository, EmployeeGroup::class.java) {
    /** Checks if a group with the given [name] exists. */
    fun existsByName(name: String): Boolean {
        return repository.existsByName(name)
    }

    /** Gets all the employees of the group with the given [id]. */
    fun getEmployeesForGroup(id: Long): Collection<Employee> {
        return employeeService.getByGroup(getById(id))
    }

    @Transactional
    override fun save(entity: EmployeeGroup): EmployeeGroup {
        fun createDefaultGroupUser(group: EmployeeGroup) {
            employeeService.save(Employee(
                    id = 1000000L + group.id!!,
                    firstName = group.name,
                    lastName = "Employee",
                    password = employeeService.passwordEncoder.encode(group.name),
                    isDefaultGroupUser = true,
                    group = group
            ))
        }

        val group = super.save(entity)
        createDefaultGroupUser(group)
        return group
    }

    override fun update(entity: EmployeeGroup): EmployeeGroup {
        val persistedGroup = getById(entity.id!!)
        with(repository.findByName(entity.name)) {
            if (this != null && id != entity.id)
                throw EntityAlreadyExistsRestException(entity.name)
        }

        return super.update(with(entity) {
            EmployeeGroup(
                    entity.id,
                    if (name.isNotBlank()) entity.name else persistedGroup.name,
                    if (permissions.isNotEmpty()) entity.permissions else persistedGroup.permissions,
                    persistedGroup.employees
            )
        })
    }

    @Transactional
    override fun delete(entity: EmployeeGroup) {
        employeeService.delete(employeeService.getDefaultGroupUser(entity))
        super.delete(entity)
    }

    /** Gets the default group cookie for the given HTTP [request]. */
    fun getRequestDefaultGroup(request: HttpServletRequest): EmployeeGroup {
        val defaultGroupCookie = WebUtils.getCookie(request, defaultGroupCookieName)
                ?: throw EntityNotFoundRestException("defaultGroup")
        val defaultGroupUser = employeeService.getById(defaultGroupCookie.value.toLong(), ignoreDefaultGroupUsers = false, ignoreSystemUsers = true)
        return defaultGroupUser.group!!
    }

    /** Sets the default group cookie for the given HTTP [response]. */
    fun setResponseDefaultGroup(groupId: Long, response: HttpServletResponse) {
        val group = getById(groupId)
        val defaultGroupUser = employeeService.getDefaultGroupUser(group)
        response.addHeader("Set-Cookie", "$defaultGroupCookieName=${defaultGroupUser.id}; Max-Age=${defaultGroupCookieMaxAge}; Path=/api; HttpOnly; Secure; SameSite=strict")
    }

    /** Adds the employee with the given [employeeId] to the group with the given [groupId]. */
    fun addEmployeeToGroup(groupId: Long, employeeId: Long) {
        addEmployeeToGroup(getById(groupId), employeeService.getById(employeeId))
    }

    /**
     * Adds a given [employee] to a given [group].
     *
     * If the [employee] is already in the [group], nothing will be done.
     * If the [employee] is already in a group, it will be removed from it.
     */
    @Transactional
    fun addEmployeeToGroup(group: EmployeeGroup, employee: Employee) {
        if (employee.group != null) removeEmployeeFromGroup(employee.group!!, employee)
        if (employee.group == group) return

        group.employees.add(employee)
        employee.group = group
        update(group)
        employeeService.update(employee)
    }

    /** Removes the employee with the given [employeeId] from the group with the given [groupId]. */
    fun removeEmployeeFromGroup(groupId: Long, employeeId: Long) =
            removeEmployeeFromGroup(getById(groupId), employeeService.getById(employeeId))

    /** Removes a given [employee] from the given [group]. */
    @Transactional
    fun removeEmployeeFromGroup(group: EmployeeGroup, employee: Employee) {
        if (employee.group == null) return

        group.employees.remove(employee)
        employee.group = null
        update(group)
        employeeService.update(employee)
    }
}

@Service
class EmployeeUserDetailsService(val employeeService: EmployeeService, val securityConfigurationProperties: SecurityConfigurationProperties) : UserDetailsService {
    override fun loadUserByUsername(username: String): UserDetails {
        try {
            return loadUserByEmployeeId(username.toLong(), true)
        } catch (ex: EntityNotFoundException) {
            throw UsernameNotFoundException(username)
        } catch (ex: EntityNotFoundRestException) {
            throw UsernameNotFoundException(username)
        }
    }

    /** Loads an [User] for the given [employeeId]. */
    fun loadUserByEmployeeId(employeeId: Long, ignoreDefaultGroupUsers: Boolean = false): UserDetails {
        val employee = employeeService.getById(employeeId, ignoreDefaultGroupUsers = ignoreDefaultGroupUsers, ignoreSystemUsers = false)
        return User(employee.id.toString(), employee.password, employee.getAuthorities())
    }
}
