package dev.fyloz.trial.colorrecipesexplorer.services;

import dev.fyloz.trial.colorrecipesexplorer.builders.MaterialBuilder;
import dev.fyloz.trial.colorrecipesexplorer.builders.MaterialTypeBuilder;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityAlreadyExistsException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityNotFoundException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.ModelException;
import dev.fyloz.trial.colorrecipesexplorer.core.model.Material;
import dev.fyloz.trial.colorrecipesexplorer.core.model.MaterialType;
import dev.fyloz.trial.colorrecipesexplorer.core.services.files.SimdutService;
import dev.fyloz.trial.colorrecipesexplorer.core.services.model.MaterialService;
import dev.fyloz.trial.colorrecipesexplorer.core.services.model.MixQuantityService;
import dev.fyloz.trial.colorrecipesexplorer.dao.MaterialDao;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockMultipartFile;
import org.springframework.test.context.junit.jupiter.SpringExtension;

import java.util.Arrays;
import java.util.List;
import java.util.Optional;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.*;

@ExtendWith(SpringExtension.class)
@SpringBootTest
public class MaterialServiceTest extends AbstractServiceTest<Material, MaterialService, MaterialDao> {

    @MockBean
    private MaterialDao materialDao;

    @Autowired
    @SpyBean
    private MaterialService materialService;

    @MockBean
    private MixQuantityService mixQuantityService;

    @MockBean
    private SimdutService simdutService;

    @BeforeEach
    void setUp() {
        materialService.setMaterialDao(materialDao);
        materialService.setMixQuantityService(mixQuantityService);
        materialService.setSimdutService(simdutService);

        super.init(materialService, materialDao);

        Material material = getEntityDifferentName();
        MaterialType materialType = getMaterialType();

        when(materialDao.existsByName(anyString())).thenReturn(false);
        when(materialDao.existsByName(material.getName())).thenReturn(true);

        when(materialDao.findByName(anyString())).thenReturn(Optional.empty());
        when(materialDao.findByName(material.getName())).thenReturn(Optional.of(material));

        when(materialDao.existsByMaterialType(any())).thenReturn(false);
        when(materialDao.existsByMaterialType(materialType)).thenReturn(true);

        when(mixQuantityService.existsByMaterial(any())).thenReturn(false);
        when(mixQuantityService.existsByMaterial(material)).thenReturn(true);
    }


    @Test
    public void whenExistsByName_thenReturnTrue() {
        // given
        Material material = getEntityDifferentName();

        // when
        boolean found = materialService.existsByName(material.getName());

        // then
        assertTrue(found);
    }

    @Test
    public void whenExistsByName_thenReturnFalse() {
        // given
        Material material = getEntity();

        // when
        boolean found = materialService.existsByName(material.getName());

        // then
        assertFalse(found);
    }

    @Test
    public void whenExistsByMaterialType_thenReturnTrue() {
        // given
        MaterialType materialType = getMaterialType();

        // when
        boolean found = materialService.existsByMaterialType(materialType);

        // then
        assertTrue(found);
    }

    @Test
    public void whenExistsByMaterialType_thenReturnFalse() {
        // given
        MaterialType materialType = new MaterialTypeBuilder().name("Teinture").build();

        // when
        boolean found = materialService.existsByMaterialType(materialType);

        // then
        assertFalse(found);
    }

    @Test
    public void whenIsLinkedToMixes_thenReturnTrue() {
        // given
        Material material = getEntityDifferentName();

        // when
        boolean found = materialService.isLinkedToMixes(material);

        // then
        assertTrue(found);
    }

    @Test
    public void whenIsLinkedToMixes_thenReturnFalse() {
        // given
        Material material = getEntity();

        // when
        boolean found = materialService.isLinkedToMixes(material);

        // then
        assertFalse(found);
    }

    @Test
    public void whenGetAllNotMixType_thenReturnMaterialList() {
        // given
        Material material = getEntity();
        Material anotherMaterial = getEntityMixType();
        List<Material> materials = Arrays.asList(material, anotherMaterial);

        doReturn(materials).when(materialService).getAll();

        // when
        List<Material> found = materialService.getAllNotMixType();

        // then
        assertFalse(found.isEmpty());
        assertTrue(found.contains(material));
        assertFalse(found.contains(anotherMaterial));
    }

    @Test
    public void whenGetByName_thenReturnMaterial() {
        // given
        Material material = getEntityDifferentName();

        // when
        Material found = materialService.getByName(material.getName());

        // then
        assertEquals(material, found);
    }

    @Test
    public void whenGetByName_thenThrowEntityNotFoundExceptionName() {
        // given
        Material material = getEntity();

        // then
        EntityNotFoundException thrown = assertThrows(EntityNotFoundException.class, () -> materialService.getByName(material.getName()));
        assertEquals(thrown.getIdentifierType(), ModelException.IdentifierType.NAME);
    }

    @Override
    @Test
    public void whenSave_thenReturnEntity() {
        // given
        Material material = getEntityDifferentId();
        MockMultipartFile simdut = new MockMultipartFile("simdut", "simdut.pdf", MediaType.APPLICATION_PDF_VALUE, "pdf content".getBytes());

        // when
        Material found = materialService.save(material, simdut);

        //then
        verify(simdutService).write(material, simdut);
        assertEquals(material, found);
    }

    @Test
    public void whenUpdate_thenReturnMaterial() {
        // given
        Material material = getEntityDifferentName();

        // when
        Material found = materialService.update(material);

        // then
        assertEquals(material, found);
    }

    @Test
    public void whenUpdate_thenThrowEntityAlreadyExistsExceptionName() {
        // given
        Material material = getEntityDifferentName();
        Material anotherMaterial = getEntityDifferentNameAndId();

        when(materialDao.findByName(anotherMaterial.getName())).thenReturn(Optional.of(anotherMaterial));

        // then
        EntityAlreadyExistsException thrown = assertThrows(EntityAlreadyExistsException.class, () -> materialService.update(material));
        assertEquals(thrown.getIdentifierType(), ModelException.IdentifierType.NAME);
    }

    @Test
    public void whenUpdate_thenSimdutServiceIsCalled() {
        // given
        Material material = getEntityDifferentName();
        MockMultipartFile simdut = new MockMultipartFile("simdut", "simdut.pdf", MediaType.APPLICATION_PDF_VALUE, "pdf content".getBytes());

        // when
        Material found = materialService.update(material, simdut);

        // then
        assertEquals(material, found);
        verify(simdutService).update(simdut, found);
    }

    @Test
    public void whenDelete_thenSimdutServiceIsCalled() {
        // given
        Material material = getEntityDifferentName();

        // when
        materialService.delete(material);

        // then
        verify(simdutService).delete(material);
    }


    @Override
    protected Material getEntity() {
        return new MaterialBuilder().build();
    }

    @Override
    protected Material getEntityDifferentId() {
        return new MaterialBuilder()
                .id(1L)
                .build();
    }

    @Override
    protected Material getEntityNullId() {
        return new MaterialBuilder()
                .id(null)
                .build();
    }

    private Material getEntityDifferentName() {
        return new MaterialBuilder()
                .name("Dirt")
                .build();
    }

    private Material getEntityDifferentNameAndId() {
        return new MaterialBuilder()
                .id(1L)
                .name("Dirt")
                .build();
    }

    private Material getEntityMixType() {
        return new MaterialBuilder()
                .mixType(true)
                .build();
    }

    private MaterialType getMaterialType() {
        return new MaterialTypeBuilder().build();
    }
}
